
#
# Copyright (C) 2002,2003 asdfGroup Inc. All rights reserved.
#
# This library is a commercial product and it is subject to the license
# agreement that can be found at the following URL:
#   http://www.asdfGroup.com/license.html
# Please contact our staff if you will have any questions:
#   mailto:staff@asdfGroup.com
#

=head1 NAME

B<POWER::SQL> - Generate basic SQL queries

=head1 SYNOPSIS

 use POWER::SQL;

 $Row	= $dbh->Select('Table', \%Q);
 @Rows	= $dbh->Select(['Table1','Table2'], {%Q, id_user=>$id});
 $count = $dbh->Count('Table', {age__gt=>25});
 $id	= $dbh->ID('Table', {login=>$login,pass=>$pass});
 $dbh->Insert('Table', \%Q);
 $dbh->InsertIgnore('Table', \%Q);
 $dbh->Update('Table', \%Q);
 $dbh->Replace('Table', \%Q);
 $dbh->Delete('Table', \%Q);
 $dbh->Delete(undef, {id_user=>$id});

 @Rows	= $dbh->All("SELECT * FROM Table WHERE id_user=?", $id);
 $Row	= $dbh->Row("SELECT * FROM Table WHERE id_user=?", $id);
 @Col	= $dbh->Col("SELECT id_user FROM Table");

=head1 DESCRIPTION

       
 SQL-      CGI.

:  __  CGI (     'Table')
  %Q,   $dbh->Select('Table', \%Q);   
   'Table' -      ;
  $dbh->Select('Table', {%Q, id_member=>$id});  
       'Table',      
  id_member  $id.

         
 SQL-   .  ,  
  POWER::SQL      DBI.
        
 POWER::SQL.

=head2  (  asdfStyle)

=over 4

=item *

       PRIMARY KEY.

=item *

    (JOIN)  ,  
-     PRIMARY KEY      .
    .
    ,      
         
 .
    ,      .

=item *

     ,    "-"  "__".

=item *

       _cdate  _mdate   
        
  $dbh->Insert(), $dbh->InsertIgnore(), $dbh->Update() 
$dbh->Replace().

=item *

   ( Site)   
     ( Member)    -
id   (.. Site.id_member).

=item *

       "__"
    (  "-").

=back

=head2     SQL-

=over 4

=item *

    "-"   :

=over 4

=item B<-order>

    ORDER BY.   -  
     : "_"  "_ ASC" 
"_ DESC".

=item B<-group>

 -order   GROUP BY.

=item B<-force>

  -force ,    Update()  Delete()
   WHERE.

=back

=item *

   "__"     "__".
  ()        .
   -      .
       .

  ,   -    
       (, ,  B<eq>
 B<ne>),          
   AND.

: { html__contain => ["<P>","<BR>"] }  
"html LIKE '%<P>%' AND html LIKE '%<BR>%'".

       WHERE.

=item *

    "".
  ()        .
   - .

: { name => "Alex" }   "name='Alex'".

       SET.

=back

=head1 METHODS

=over 4

=cut

package POWER::SQL;

use strict;

our $VERSION = '1.51';
our $dbh;

sub dbq { join ',', map { $dbh->quote($_) } ref $_[0] ? @{$_[0]} : $_[0] }
sub is_int {$_[0] =~ m/^-?\d+ +(?:SECOND|MINUTE|HOUR|DAY|MONTH|YEAR)$/}
my %Func = (
    # FIXME ,     eq  ne   
    # _[cm]date  Insert(), InsertIgnore(), Update()  Replace()
    # FIXME   eq/ne    SET (..  {FieldList}),
    #    'name IS NULL'   'name = NULL' (
    #     P2SQL())
    # BUG  'name IS NULL'  auto_increment  
    # LAST_INSERT_ID(),    auto_increment   workaround
    'eq'        => sub {
		       my (@val, $null, @expr);
		       @val  = ref $_[0] ? @{$_[0]} : $_[0];
		       $null = 0 + grep {!defined} @val;
		       @val  = grep {defined} @val;
		       push @expr, "$_[1] IS NULL"		   if $null;
		       push @expr, "$_[1] = ".dbq(\@val)	   if @val==1;
		       push @expr, "$_[1] IN (".dbq(\@val).")"	   if @val>1;
		       push @expr, 'NOT 1'			   unless @expr;
		       @expr==1 ? $expr[0] : '('.join(' OR ', @expr).')';
		   },
    'ne'        => sub {
		       my (@val, $null, @expr);
		       @val  = ref $_[0] ? @{$_[0]} : $_[0];
		       $null = 0 + grep {!defined} @val;
		       @val  = grep {defined} @val;
		       push @expr, "$_[1] IS NOT NULL" if $null and not @val;
		       push @expr, "$_[1] IS NULL"     if not $null and @val;
		       push @expr, "$_[1] != ".dbq(\@val)	   if @val==1;
		       push @expr, "$_[1] NOT IN (".dbq(\@val).")" if @val>1;
		       push @expr, 'NOT 0'	       unless @expr;
		       @expr==1 ? $expr[0] : '('.join(' OR ', @expr).')';
		   },
    'lt'        => sub { "$_[1] <  ".&dbq },
    'gt'        => sub { "$_[1] >  ".&dbq },
    'le'        => sub { "$_[1] <= ".&dbq },
    'ge'        => sub { "$_[1] >= ".&dbq },
    'setdate'   => sub { &is_int and "$_[1] =  DATE_ADD(NOW(), INTERVAL $_[0])" },
    'date_eq'   => sub { &is_int and "$_[1] =  DATE_ADD(NOW(), INTERVAL $_[0])" },
    'date_ne'   => sub { &is_int and "$_[1] != DATE_ADD(NOW(), INTERVAL $_[0])" },
    'date_lt'   => sub { &is_int and "$_[1] <  DATE_ADD(NOW(), INTERVAL $_[0])" },
    'date_gt'   => sub { &is_int and "$_[1] >  DATE_ADD(NOW(), INTERVAL $_[0])" },
    'date_le'   => sub { &is_int and "$_[1] <= DATE_ADD(NOW(), INTERVAL $_[0])" },
    'date_ge'   => sub { &is_int and "$_[1] >= DATE_ADD(NOW(), INTERVAL $_[0])" },
    'like'      => sub { "$_[1] LIKE ".&dbq },
    'notlike'   => sub { "$_[1] NOT LIKE ".&dbq },
    'contain'   => sub { "$_[1] LIKE ".dbq('%'.$_[0].'%') },
    'notcontain'=> sub { "$_[1] NOT LIKE ".dbq('%'.$_[0].'%') },
    'inc'       => sub { "$_[1] = $_[1] + ".&dbq },
    );
my %DESC;

=item B<P2SQL>( $table, \%P )

=item B<P2SQL>( \@tables, \%P )

 $table  @tables   ,    ,  %P
  ,       $table  @tables.
   ( {ID}, {Field}, ...)    "."
    .

   :

=over 4

=item {DB}

       ( TEMPORARY)

=item {Desc}

       
  "" => $dbh->All("DESC ")

=item {Table}

   

=item {INNER}

   INNER JOIN

=item {LEFT}

   LEFT JOIN

=item {ID}

  PRIMARY KEY  {Table}

=item {FieldList}

  "=, 2=2"  
%P   ""

=item {FieldAND}

a-la {FieldList},     "AND" 
 -    "__"
        "1"

=item {Update}

a-la {FieldAND},      "" 
 primary key,   .

=item {Select}

     "SELECT *"  
   :  SELECT   
   "field",   {Select}  
    : "..., table1.field, ...".
 {Select}       `, '.

=item {Order}

  " ASC, 2 DESC"  ""

=item {Group}

a-la {Order}

=back

=cut

sub DBI::db::P2SQL {
    (local $dbh, my $tables, my $P) = @_;
    my %SQL = map {$_=>''} qw(Select FieldList FieldAND Update);
    #        
    #  $SQL{DB}
    $DESC{$dbh} ||= { map {$_=>undef} $dbh->Col('SHOW TABLES') };
    $SQL{DB} = [ keys %{$DESC{$dbh}} ];
    #      ,     
    my $was_tables = ref $tables ? 0+@$tables : defined $tables ? 1 : 0;
    my @tables = ref $tables ? @$tables : defined $tables ? $tables : ();
    #      
    for (grep {not $DESC{$dbh}{$_}} @tables) {
	next unless my @desc = $dbh->All("DESC $_");
        $DESC{$dbh}{$_} ||= \@desc;
    }
    @tables = grep { exists $DESC{$dbh}{$_} and @{$DESC{$dbh}{$_}} } @tables;
    die 'Unknown table in ['.join(',',ref $tables ? @$tables : $tables).']'
	if $was_tables != @tables;
    # ,       (..    )
    return %SQL unless @tables;
    #  $SQL{Table}
    $SQL{Table} = $tables[0];
    #  $SQL{Desc}   
    $SQL{Desc}{ $SQL{Table} } = [map{{%$_}}@{   $DESC{$dbh}{ $SQL{Table} }   }];
    #  $SQL{ID}
    $SQL{ID} = $SQL{Desc}{ $SQL{Table} }[0]{Field};
    #  ,     
    my @join = $tables[0];
    my @field = map {{ map {$_->{Field}=>1} @{$DESC{$dbh}{$_}} }} @tables;
    for my $right (1..$#tables) {
        my $rkey = $DESC{$dbh}{ $tables[$right] }[0]{Field};
        for my $left (0..$right-1) {
            my $lkey = $DESC{$dbh}{ $tables[$left] }[0]{Field};
            my $key = $field[$left]{$rkey}  ? $rkey : 
                      $field[$right]{$lkey} ? $lkey : next;
            push @join, "JOIN $tables[$right]".
                " ON ($tables[$right].$key=$tables[$left].$key)";
            #  $SQL{Desc}    
            $SQL{Desc}{ $tables[$right] } = [map{{%$_}}@{
		$DESC{$dbh}{ $tables[$right] }   }];
            last;
        }
    }
    die 'Can\'t join tables: ['.join(',',ref $tables ? @$tables : $tables).']'
	if $was_tables != @join;
    #  $SQL{INNER}  $SQL{LEFT}
    $SQL{INNER} = join ' INNER ', @join;
    $SQL{LEFT}  = join ' LEFT ',  @join;
    #       
    #  $SQL{Select}    
    my %qualify;
    for my $t (@tables) {
        for (map {$_->{Field}} @{$SQL{Desc}{$t}}) {
            next if $qualify{$_};
            $qualify{$_} = "$t.$_";
            $SQL{Select} .= ", $t.$_";
        }
    }
    $SQL{Select} =~ s/^, //;
    #  $SQL{FieldList}, $SQL{FieldAND}, $SQL{Update}
    for my $f (keys %$P) {
        my ($name, $func) = map {$_||''} $f =~ /^(.*?)(?:__(.*))?$/;
        next unless my $qname = $qualify{$name};            #   
        die "Unknown function: $func"
	    unless my $sub   = $Func{$func || 'eq'};	    #   
	die "Array ref without function: $f" 
	    if ref($P->{$f}) and ! $func;		    # field=>[1,2,3]
	#  `eq'  `ne'   array ref :
        my $is_list = ref($P->{$f}) && ($func ne 'eq') && ($func ne 'ne');
      for my $value ($is_list ? @{$P->{$f}} : $P->{$f}) {
        die "Bad value [$value] for function [$func]"
	  unless my $expr  = $sub->($value, $qname);	    # . .
        $SQL{FieldList} .= ", $expr" if (!$func or $func eq 'inc' or
            $func eq 'setdate');
        $SQL{FieldAND}  .= " AND $expr" if ($func ne 'inc' and 
	    $func ne 'setdate');
        $SQL{Update}    .= " AND $expr" if (($func and $func ne 'inc' and
            $func ne 'setdate') or $name eq $SQL{ID});
      }
    }
    $SQL{FieldList} =~ s/^, //;
    $SQL{FieldAND}  =~ s/^ AND //; $SQL{FieldAND} ||= '1';
    $SQL{Update}    =~ s/^ AND //; $SQL{Update}   ||= '1';
    $SQL{FieldList} =~ s/\s+IS\s+NULL/ = NULL/g;
    #  $SQL{Order}
    for my $order (ref $P->{-order} ? @{$P->{-order}} : $P->{-order}) {
	next unless defined $order;
        die "Bad order format: $order" unless $order =~ /^(\w+)( ASC| DESC|)$/;
        die "Bad order field: $order"  unless $qualify{$1};
        $SQL{Order} = join ', ', ($SQL{Order}||()), $qualify{$1}.$2;
    }
    #  $SQL{Group}
    for my $group (ref $P->{-group} ? @{$P->{-group}} : $P->{-group}) {
	next unless defined $group;
        die "Bad group format: $group" unless $group =~ /^(\w+)( ASC| DESC|)$/;
        die "Bad group field: $group"  unless $qualify{$1};
        $SQL{Group} = join ', ', ($SQL{Group}||()), $qualify{$1}.$2;
    }
    return %SQL;
}

=item B<Insert>( $table, \%P )

INSERT INTO ?table SET ?set

     "_cdate",  "_mdate"     NOW().
 $dbh->{mysql_insertid}     undef().

=cut

sub DBI::db::Insert {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {%$P, '_cdate'=>undef, '_mdate'=>undef});
    die 'no data to Insert()' unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_[cm]date\s*)=\s*NULL/$1=NOW()/g;
    return $dbh->do("INSERT INTO $SQL{Table} SET $SQL{FieldList}")
        ? $dbh->{mysql_insertid} : undef;
}

=item B<InsertIgnore>( $table, \%P )

INSERT IGNORE INTO ?table SET ?set

     "_cdate",  "_mdate"     NOW().
      undef().
   LAST_INSERT_ID()  INSERT IGNORE !

=cut

sub DBI::db::InsertIgnore {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {%$P, '_cdate'=>undef, '_mdate'=>undef});
    die 'no data to Insert()' unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_[cm]date\s*)=\s*NULL/$1=NOW()/g;
    return $dbh->do("INSERT IGNORE INTO $SQL{Table} SET $SQL{FieldList}");
}

=item B<Update>( $table, \%P )

UPDATE ?table SET ?set WHERE ?id=... AND ?where

     "_mdate",  '_mdate=NOW()'.
   SET     "",    WHERE  
  "__"  primary_key     "".
 -    undef()  .

     WHERE  {-force=>1}  %P.

=cut

sub DBI::db::Update {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {%$P, '_mdate'=>undef});
    die 'no data to Update()' unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_mdate\s*)=\s*NULL/$1=NOW()/g;
    die 'Update() with empty WHERE require {-force=>1}'
	if $SQL{Update} eq '1' and not $P->{-force};
    return $dbh->do("UPDATE $SQL{Table} SET $SQL{FieldList} WHERE $SQL{Update}");
}

=item B<Replace>( $table, \%P )

REPLACE ?table SET ?set

     "_cdate"  "_mdate",     NOW().
   SET     "".
    REPLACE  undef().

=cut

sub DBI::db::Replace {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, {%$P, '_cdate'=>undef, '_mdate'=>undef});
    die 'no data to Replace()' unless $SQL{Table} and $SQL{FieldList};
    $SQL{FieldList} =~ s/(_[cm]date\s*)=\s*NULL/$1=NOW()/g;
    return $dbh->do("REPLACE INTO $SQL{Table} SET $SQL{FieldList}");
}

=item B<Delete>( $tables, \%P )

DELETE FROM ?table WHERE ?where

    .  $tables  undef(),  
    ( TEMPORARY)      %P .

     WHERE  {-force=>1}  %P.

 -    $dbh->do().    
,   -     .

=cut

sub DBI::db::Delete {
    my ($dbh, $table, $P) = @_;
    if (!defined $table) {
        die 'can\'t Delete() without table and params'
	    unless my %field = map {s/__.*//;$_=>1} keys %$P;
        for my $t (@{{$dbh->P2SQL()}->{DB}}) {
            my %seen;
            push @$table, $t if keys(%field) == grep {++$seen{$_}==2} 
                keys(%field), map {$_->{Field}} @{{$dbh->P2SQL($t)}->{Desc}{$t}};
        }
    }
    my $res;
    for my $t (ref $table ? @$table : $table) {
        my %SQL = $dbh->P2SQL($t, $P);
        die 'no table for Delete()' unless $SQL{Table};
	die 'Delete() with empty WHERE require {-force=>1}'
	    if $SQL{FieldAND} eq '1' and not $P->{-force};
        $res = $dbh->do("DELETE FROM $SQL{Table} WHERE $SQL{FieldAND}");
    }
    return $res;
}

=item B<ID>( $tables, \%P )

SELECT primary_key FROM ?inner WHERE ?where [ORDER BY ?order]

   Col() .

=cut

sub DBI::db::ID {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, $P);
    die 'no table for ID()' unless $SQL{Table};
    $dbh->Col("SELECT $SQL{Table}.$SQL{ID} FROM $SQL{INNER}".
	" WHERE $SQL{FieldAND}".($SQL{Order} ? " ORDER BY $SQL{Order}" : ''));
}

=item B<Count>( $tables, \%P )

SELECT count(*) FROM ?inner WHERE ?where

   Col() .

=cut

sub DBI::db::Count {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, $P);
    die 'no table for Count()' unless $SQL{Table};
    $dbh->Col("SELECT count(*) __count FROM $SQL{INNER} WHERE $SQL{FieldAND}");
}

=item B<Select>( $tables, \%P )

SELECT * FROM ?inner WHERE ?where [ORDER BY ?order]
SELECT *, count(*) __count FROM ?inner WHERE ?where GROUP BY ?group
    [ORDER BY ?order]

 "SELECT *"       
  ;       
     -      .

   All()      Row()  .

=cut

sub DBI::db::Select {
    my ($dbh, $table, $P) = @_;
    my %SQL = $dbh->P2SQL($table, $P);
    die 'no table for Select()' unless $SQL{Table};
    my $SQL = 'SELECT '.$SQL{Select}.
        ($SQL{Group} ? ', count(*) __count' : '').
        " FROM $SQL{INNER} WHERE $SQL{FieldAND}".
        ($SQL{Group} ? " GROUP BY $SQL{Group}" : '').
        ($SQL{Order} ? " ORDER BY $SQL{Order}" : '');
    return wantarray ? $dbh->All($SQL) : $dbh->Row($SQL);
}

=item B<All>($sql[, @bind])

Execute $sql and return array of hashrefs to row data.
Example:

 [ { id => 1, user => "root" },
   { id => 2, user => "daemon" },
   { id => 3, user => "system" },
 ]

=cut

sub DBI::db::All {
    my ($dbh, $sql, @bind) = @_;
    (my $sth = $dbh->prepare($sql))->execute(@bind);
    @{ $sth->fetchall_arrayref({}) };
}

=item B<Row>($sql[, @bind])

Execute $sql and return reference to hash with row data or undef.
Example:

 { id => 1, user => "root" }

=cut

sub DBI::db::Row {
    my ($dbh, $sql, @bind) = @_;
    (my $sth = $dbh->prepare($sql))->execute(@bind);
    $sth->fetchrow_hashref();
}

=item B<Col>($sql[, @bind])

Execute $sql and return array of column data in list context or
first cell in scalar context.

=cut

sub DBI::db::Col {
    my ($dbh, $sql, @bind) = @_;
    my @res = @{ $dbh->selectcol_arrayref($sql, undef, @bind) || [] };
    return wantarray ? @res : $res[0];
}

=back

=head1 FUNCTIONS for FIELDS

,         "__":

=over 8

=item B<eq>, B<ne>, B<lt>, B<gt>, B<le>, B<ge>

name = != < > <= >= value

   B<eq>  B<ne>:

 eq []		 - NOT 1
 ne []		 - NOT 0
 eq  undef - name IS NULL
 ne  undef - name IS NOT NULL
 eq 	   undef - name IN (...)
 ne     undef - (name IS NULL OR name NOT IN (...))
 eq       undef - (name IS NULL OR name IN (...))
 ne       undef - name NOT IN (...)



 "[]"	       : name__func=>[]
 " undef": name__func=>undef     name__func=>[undef]
 "    undef": name__func=>$defined  name__func=>[@defined]
 "      undef": name__func=>[@defined_and_not_defined]

=item B<date_eq>, B<date_ne>, B<date_lt>, B<date_gt>, B<date_le>, B<date_ge>

name = != < > <= >= DATE_ADD(NOW(), INTERVAL value)

value must be /^-?\d+ (?:SECOND|MINUTE|HOUR|DAY|MONTH|YEAR)$/

=item B<like>, B<notlike>

name [NOT] LIKE value

=item B<contain>, B<notcontain>

name [NOT] LIKE '%'.value.'%'

=item B<inc>

name = name + value

  Update()    SET,   WHERE.
     Insert(), InsertIgnore()  Replace()
.

=item B<setdate>

name = DATE_ADD(NOW(), INTERVAL value)

  Insert(), InsertIgnore(), Update()  Replace() 
   SET.

=back

=head1 BUGS

  PRIMARY KEY   :

 { id => $new_id, id__eq => $old_id }

..  id  id__eq    WHERE:

 SET id = $new_id WHERE id = $new_id AND id = $old_id

    'id => $new_id'   WHERE 
 'id__eq',       CGI 
'($table, {%Q, id_member=>$S{id_member}, id_member__eq=>$S{id_member})'
     id_member  id_member__eq
     id_member     
id_member.

=head1 HISTORY

 1.51	      20 16:59:57 EET 2005
	- InsertIgnore()    LAST_INSERT_ID()
	  .. MySQL     
	  INSERT IGNORE

 1.50	       8 17:58:01 EET 2005
	- added InsertIgnore()
	- SELECT LAST_INSERT_ID() replaced by $dbh->{mysql_insertid}

 1.42	       8 00:25:24 EET 2005
	- minor fix to support 'use strict;'

 1.41	      26 16:49:52 EEST 2004
	- bugfix in ID()

 1.40	      17 19:36:12 EEST 2003
	- __eq  __ne      
	       undef
	WARNING INCOMPATIBLE CHANGES:
	     field__ne=>["Y"]  -
	     field IS NULL
	-     BUGS

 1.33	      25 01:54:32 EEST 2003
	- minor fixes for `perl -w` mode

 1.32	      11 04:31:49 EET 2002
	- Delete()   -  
	     $dbh->do()
	  (    - 1.00)
	-       {DB}  ''
	  (    - 1.31)
	-  P2SQL()     
	     %DESC,     
	   

 1.31	      10 18:11:22 EET 2002
	-  TEMPORARY 

 1.30	      15 05:15:18 EEST 2002
	- :       
	    ""
	- :   -force   Update()
	   Delete()    WHERE
	-       die()
	-   
	-  pod 
	- minor bugfix:  __inc    WHERE  Delete() 
	   Select()

 1.20         16 12:27:53 EEST 2002
        -  
        -  

 1.11          2 18:45:41 EEST 2002
        -  {FieldAND}  {Update}  = NULL  != NULL
            IS NULL  IS NOT NULL

 1.10         30 23:53:42 EET 2002
        -   {Select}  P2SQL
        -  Select()     "*",   {Select},
            ,      
             ,   
               .

 1.08         30 16:52:13 EET 2002
        -   __B<setdate>   
           NOW()  Insert()  Update()

 1.07         27 02:25:18 EET 2002
        -   "not"   "!" -   
          perl 5.005_03

 1.06         25 15:54:53 EET 2002
        -   __B<eq>  __B<ne>   
                 
        -   -     
             Delete(undef, $P)
        -        -order  -group
            

 1.05         24 01:31:32 EET 2002
        -      _cdate  _mdate

 1.04         23 16:24:16 EET 2002
        -   -     
              

 1.03         22 16:48:19 EET 2002
        -   __B<inc>

 1.02         20 13:50:16 EET 2002
        -   __B<btw>
        -  IN   __B<eq>  __B<ne>
        -         __B<eq> 
           __B<ne>      
            

 1.01         17 23:19:20 EET 2002
        - bugfix:   -order  -group  ASC|DESC
        - bugfix:   -order  -group 
            
        -   __B<like>  __B<notlike>
        -   __B<contain>  __B<notcontain>

 1.00         15 16:52:30 EET 2002
        - first release

=head1 AUTHORS

 Alex Efros <powerman@asdfGroup.com>
 Nikita Savin <nikita@asdfGroup.com>

 Copyright (C) 2002 asdfGroup Inc. All rights reserved.

 This library is a commercial product and it is subject to the license
 agreement that can be found at the following URL:
    http://www.asdfGroup.com/license.html
 Please contact our staff if you will have any questions:
    mailto:staff@asdfGroup.com

=cut

1;
